/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "wresample.hh"
#include "matlab_fcs.hh"
#include "lcl_array.hh"
#include "Bits.hh"
//#include "DecimateBy2.hh"
#include "Hanning.hh"
#include "DVecType.hh"
#include "FIRdft.hh"
#include "FIRdesign.hh"
#include <iostream>

//--------------------------------------------------------------------------

using namespace std;

namespace wpipe {

   resampler::filt_map resampler::rate_map;

   //======================================  Resampler contructor
   resampler::resampler(void) {}

  //======================================  Resampler destructor
  resampler::~resampler(void) {
    reset();
    for (filt_iter i=rate_map.begin(); i != rate_map.end(); ++i) {
      delete i->second;
      i->second = 0;
    }
    rate_map.clear();
  }

  //======================================  Resample function
  tser_vect
  resampler::wresample(const tser_vect& data, double sampleFrequency) {
     size_t numberOfChannels = data.size();
     if (pipe_vect.size() != numberOfChannels) reset();
     if (pipe_vect.empty()) pipe_vect.resize(numberOfChannels, 0);
     if (out_data.empty())  out_data.resize(numberOfChannels);

    ///////////////////////////////////////////////////////////////////////////
    //                   validate command line arguments                     //
    ///////////////////////////////////////////////////////////////////////////
 
    //==================================   validate equal data durations
     Interval dataDuration = data[0].getInterval();
     for (size_t chanNumber=1; chanNumber < numberOfChannels; chanNumber++) {
	if (dataDuration != data[chanNumber].getInterval()) {
	   error("data durations are not equal");
	}
     }

     //////////////////////////////////////////////////////////////////////////
     //                            resample data                             //
     //////////////////////////////////////////////////////////////////////////

     // begin loop over channels
     tser_vect ts(numberOfChannels);
     long fSample = long(sampleFrequency);
     for (size_t chanNumber=0; chanNumber<numberOfChannels; chanNumber++){
	TSeries tsin(data[chanNumber]);
	tsin.Convert(DVector::t_double);

	// for now demand reduction by a power of 2
	double decimFact = sampleFrequency*double(tsin.getTStep());
	if (decimFact == 0) {
	   error("Decimation factor is zero");
	}
	decimFact = 1.0/decimFact;
	if (!is_power_of_2(int(decimFact))) {
	   cerr << "Target rate: " << sampleFrequency << " input t-step: " 
		<< tsin.getTStep() << endl;
	   error("Sample rate reduction not a power of 2");
	}

	long fRawSample = long(1./double(tsin.getTStep()));
	if (fSample != fRawSample) {
	   // resample data
	   //ts = resample(data, upSampleFactor, downSampleFactor, filter)

	   //------------------------------  Get an anti-aliasing filter 
	   //                                Note: assume integer sample rates
	   if (!pipe_vect[chanNumber]) {
	      pipe_vect[chanNumber] = filter(fRawSample, fSample);
	   }

	   FIRdft& filt = dynamic_cast<FIRdft&>(*pipe_vect[chanNumber]);
	   Interval tShift = filt.getTimeDelay();
	   Time tFilt = filt.getCurrentTime();
	   if (!tFilt || tFilt < tsin.getStartTime()) {
	      filt.reset();
	      out_data[chanNumber].Clear();
	   }
	   Time tInput = tsin.getStartTime();
	   Time tOutput = out_data[chanNumber].getStartTime();
	   if (tOutput != Time(0) && tOutput < tInput) {
	      out_data[chanNumber].eraseStart(tOutput - tInput);
	   }
	   if (tFilt > tInput) {
	      tsin.eraseStart(tFilt - tInput);
	   }

	   //------------------------------  Filter and decimate
	   long downSampleFactor = fRawSample / fSample;
	   out_data[chanNumber].Append(filt(tsin).decimate(downSampleFactor));

	   //  On first stride, remove data from before nominal start time.
	   if (!tOutput) {
	      out_data[chanNumber].eraseStart(tInput - tOutput);
	   }
	   tsin = out_data[chanNumber];
	   tsin.extend(data[chanNumber].getEndTime());
	}
	ts[chanNumber] = tsin;
     }      // end loop over channels

     //////////////////////////////////////////////////////////////////////////
     //                                  return                              //
     //////////////////////////////////////////////////////////////////////////

     // return to calling function
     return ts;
  }

  Pipe*
  resampler::filter(long fRawSample, long fSample) {
    rate_pair rp(fRawSample, fSample);
    if (rate_map.find(rp) == rate_map.end()) {
      long denom      = gcd(fSample, fRawSample);
      long upSampleFactor   = fSample /denom;
      long downSampleFactor = fRawSample / denom; 

      // design anti-alias filter
      const int nBand = 2;
      int filterOrder = 2 * 256 * max(upSampleFactor, downSampleFactor);
      double filterCutoff = 0.99 / max(upSampleFactor, downSampleFactor);
      double filterFrequencies[2*nBand] = {0.0, filterCutoff, filterCutoff, 1.0};
      double filterMagnitudes[2*nBand]  = {1.0, 1.0, 0.0, 0.0};
      double wband[nBand] = {1.0, 1.0};
      DVectD filter_coefs(filterOrder+1);
      firls(filterOrder, nBand, filterFrequencies, filterMagnitudes, wband, 
	    filter_coefs.refTData());
      Hanning filter1(filterOrder + 1);
      filter_coefs *= filter1.refDVect();
      filter_coefs *= double(upSampleFactor)
	             /filter1.refDVect().getDouble(filterOrder/2);
      FIRdft* filt = new FIRdft(filterOrder, double(fRawSample));
      filt->setCoefs(filterOrder+1, filter_coefs.refTData());
      filt->setMode(FIRdft::fm_zero_phase);
      rate_map.insert(filt_node(rp, filt));
    }
    return rate_map.find(rp)->second->clone();
  }

  //====================================  Reset all filters.
  void
  resampler::reset(void) {
     size_t N = pipe_vect.size();
     for (size_t i=0; i<N; ++i) {
	delete pipe_vect[i];
	pipe_vect[i] = 0;
     }
     pipe_vect.clear();
  }

}  // namespace wpipe
